#=======================================================================
#	$Id: Model.pm,v 1.7 2006/08/10 15:23:59 pythontech Exp $
#	Model based on triple store.
#	Copyright (C) 2004-2006  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#-----------------------------------------------------------------------
#	In the world seen by users of the model, there are:
#	  PythonTech::Obj (or subclass)		resource (URI = $obj->uri)
#	  \$uri					resource
#	  scalar				literal
#=======================================================================
package PythonTech::Model;
use strict;
use PythonTech::Obj;

#--- Predefined qname prefixes for standard stuff
#my $rdf = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#';
#my $rdfs = 'http://www.w3.org/2000/01/rdf-schema#';
my $rdf = 'rdf:';

#-----------------------------------------------------------------------
#	Create new model.  Configuration options may be given either
#	as a hash or as many name=>value arguments
#	  triples	Triple store e.g. PythonTech::Triples::DBHash
#	  classmap	Given a resource, lookup this map to find a
#			class to bless it into.  The map entry with
#			key '' is the fallback default.  If even that
#			is undefined (or no classmap is given), it
#			is blessed into PythonTech::Obj
#	  extqmap	Qname mapping for external URIs
#	  dbqmap	Qname mapping within triplestore
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {
	triples => {},		# PythonTech::Triples
	classmap => {},		# { $rdftype => $perlclass }
	objmap => {},		# { $uri => $obj }
    };
    bless $self, $class;
    $self->configure(@_);
    return $self;
}

sub configure {
    my $self = shift;
    my @props = ref($_[0]) eq 'HASH' ? %{$_[0]} : @_;
    while (my($key,$value) = splice(@props,0,2)) {
	if ($key eq 'triples') {
	    $self->{'triples'} = $value;
	} elsif ($key eq 'dbqmap') {
	    $self->{'dbqmap'} = $value;
	} elsif ($key eq 'extqmap') {
	    $self->{'extqmap'} = $value;
	} elsif ($key eq 'classmap') {
	    keys %$value;
	    while (my($class,$perlclass) = each %$value) {
		$self->{'classmap'}->{$class} = $perlclass;
	    }
	} else {
	    die ref($self).": Unknown configuration option $key\n";
	}
    }
}

#-----------------------------------------------------------------------
#	Return all resources with given RDF type
#-----------------------------------------------------------------------
sub all_of_type {
    my($self, $type) = @_;
    return $self->po2s('rdf:type',\$type);
}

#-----------------------------------------------------------------------
#	Given a resource URI, attempt to instantiate it as an object.
#	If it has already been seen, it is found in the objmap.
#-----------------------------------------------------------------------
sub as_obj {
    my($self, $uri) = @_;
#    print STDERR "as_obj: $uri\n";
    # See if already known
    my $obj = $self->{'objmap'}->{$uri};
#    if (defined($obj)) {
#	my $ouri = $obj->uri;
#	print STDERR "as_obj(KNOWN): $uri, $ouri\n";
#    }
    return $obj if defined($obj);

    # Check if it is defined (i.e. has an rdf:type)
    my $turi = $uri;
    if ($self->{'extqmap'}) {
	$turi = $self->{'extqmap'}->decode($turi);
    }
    if ($self->{'dbqmap'}) {
	$turi = $self->{'dbqmap'}->encode($turi);
    }
#    print STDERR "as_obj(Q): $q_uri\n";
    my @types = $self->{'triples'}->sp2o($turi, $rdf.'type');
#    die "$uri has no rdf:type\n" unless @types;

    # Create it in base class
    $self->{'objmap'}->{$uri} = $obj = {
	uri => $uri,
	model => $self,
    };
    bless $obj, 'PythonTech::Obj';

    # Bless it into the appropriate Perl class.
    # There may be more than one (subclassing) - make sure it is in the
    # most specific. (Beware - random choice if multiple inheritance!)
#    print STDERR "Before class blessing\n";
    foreach my $type ('',@types) { # why did it ever work ? chah
#	print STDERR "as_obj(blessing) - $uri $type\n" if $uri =~ /wiki/;
#	print STDERR "as_obj(blessing) - $uri $type\n";
	if (defined(my $perlclass = $self->{'classmap'}->{$type})) {
#	    print STDERR "as_obj - perlclass $perlclass\n" if $uri =~ /wiki/;
#	    print STDERR "as_obj - perlclass $perlclass\n";
	    unless ($obj->isa($perlclass)) {
		eval "use $perlclass; 1"
		  or die "Cannot find perl class $perlclass: $@\n";
		bless $obj, $perlclass;
#		print STDERR "as_obj - blessed as $perlclass\n" if $uri =~ /wiki/;
#		print STDERR "as_obj - blessed as $perlclass\n";
	    }
	}
    }
#    print STDERR "After class blessing\n";
    return $obj;
}

#-----------------------------------------------------------------------
#	Create a new object e.g.
#	$chah = $model->create_obj('c:user/chah','cs:Person',
#				   'cs:userName' => "chah",
#				   'cs:userBoss' => $fsg);
#-----------------------------------------------------------------------
sub create_obj {
    my($self, $uri,$type,@attr) = @_;
    $self->add(\$uri,'rdf:type',\$type);
    while (my($pred,$obj) = splice(@attr,0,2)) {
	$self->add(\$uri,$pred,$obj);
    }
    return $self->as_obj($uri);
}

#-----------------------------------------------------------------------
#	Add a statement to the model
#-----------------------------------------------------------------------
sub add {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);
    $self->{'triples'}->add($tsubj,$tpred,$tobj);
}

#-----------------------------------------------------------------------
#	Remove a statement from the model
#-----------------------------------------------------------------------
sub remove {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);
    $self->{'triples'}->remove($tsubj,$tpred,$tobj);
}

#-----------------------------------------------------------------------
#	Test if statement is true
#-----------------------------------------------------------------------
sub test {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);
    return $self->{'triples'}->test($tsubj,$tpred,$tobj);
}

#-----------------------------------------------------------------------
#	Get object(s) given subject and predicate
#-----------------------------------------------------------------------
sub sp2o {
    my($self, $subj,$pred) = @_;
    my $tsubj = $self->_insub($subj);
    my $tpred = $self->_inpred($pred);
#    print STDERR "sp2o: subj=$subj tsubj=$tsubj pred=$pred tpred=$tpred\n";
    if (wantarray) {
	my @tobjs = $self->{'triples'}->sp2o($tsubj,$tpred);
	return map {$self->_out($_)} @tobjs;
    } else {
	my $tobj = $self->{'triples'}->sp2o($tsubj,$tpred);
	return defined($tobj) ? $self->_out($tobj) : undef;
    }
}

#-----------------------------------------------------------------------
#	Get subject(s) given predicate and object
#-----------------------------------------------------------------------
sub po2s {
    my($self, $pred,$obj) = @_;
    my $tpred = $self->_inpred($pred);
    my $tobj = $self->_in($obj);

    if (wantarray) {
	my @tsubjs = $self->{'triples'}->po2s($tpred,$tobj);
	return map {$self->_out($_)} @tsubjs;
    } else {
	my $tsubj = $self->{'triples'}->po2s($tpred,$tobj);
	return defined($tsubj) ? $self->_out($tsubj) : undef;
    }
}

#-----------------------------------------------------------------------
#	Query
#-----------------------------------------------------------------------
sub query {
    my($self, $subj,$pred,$obj) = @_;
    my $tsubj = defined $subj ? $self->_insub($subj) : undef;
    my $tpred = defined $pred ? $self->_inpred($pred) : undef;
    my $tobj  = defined $obj  ? $self->_in($obj)  : undef;
    if (wantarray) {
	my @res = $self->{'triples'}->query($tsubj,$tpred,$tobj);
	return map {[map {$self->_out($_)} @$_]} @res;
    } else {
	my $res = $self->{'triples'}->query($tsubj,$tpred,$tobj);
	return defined($res) ? [map {$self->_out($_)} $res] : undef;
    }
}

#-----------------------------------------------------------------------
#	Commit any changes
#-----------------------------------------------------------------------
sub commit {
    my($self) = @_;
    $self->{'triples'}->commit;
}

#-----------------------------------------------------------------------
#	... or not
#-----------------------------------------------------------------------
sub rollback {
    my($self) = @_;
    $self->{'triples'}->rollback;
}

#-----------------------------------------------------------------------
#	Disconnect / reconnect from underlying triplestore
#-----------------------------------------------------------------------
sub disconnect {
    my $self = shift;
    $self->{'triples'}->disconnect;
}

sub reconnect {
    my $self = shift;
    $self->{'triples'}->reconnect;
}

#-----------------------------------------------------------------------
#	Convert subject to URI
#-----------------------------------------------------------------------
sub _insub {
    my($self, $value) = @_;
    my $uri;
    if (ref($value) eq 'SCALAR') {
	$uri = $$value;
    } elsif (ref($value)) {
	$uri = eval {$value->uri};
	unless ($uri) {
	    die "no uri: $value ".ref($value)."";
	}
    } else {
	$uri = $value;
    }
    my $quri = $uri;
    if ($self->{'extqmap'}) {
	$quri = $self->{'extqmap'}->decode($quri);
    }
    if ($self->{'dbqmap'}) {
	$quri = $self->{'dbqmap'}->encode($quri);
    }
    return $quri;
}

#-----------------------------------------------------------------------
#	Convert predicate to database standard URI
#-----------------------------------------------------------------------
sub _inpred {
    my($self, $uri) = @_;
    my $turi = $uri;
    if ($self->{'extqmap'}) {
	$turi = $self->{'extqmap'}->decode($turi);
    }
    if ($self->{'dbqmap'}) {
	$turi = $self->{'dbqmap'}->encode($turi);
    }
    return $turi;
}

#-----------------------------------------------------------------------
#	Convert object or literal to URI or =literal
#-----------------------------------------------------------------------
sub _in {
    my($self, $value) = @_;
    my $uri;
    if (ref($value) eq 'SCALAR') {
	$uri = $$value;
    } elsif (ref($value)) {
	$uri = eval {$value->uri};
	unless ($uri) {
	    die "no uri: $value ".ref($value)."";
	}
    } else {
	return "=$value";
    }
    my $quri = $uri;
    if ($self->{'extqmap'}) {
	$quri = $self->{'extqmap'}->decode($quri);
    }
    if ($self->{'dbqmap'}) {
	$quri = $self->{'dbqmap'}->encode($quri);
    }
    return $quri;
}

#-----------------------------------------------------------------------
#	Convert value from triple store to object or literal
#-----------------------------------------------------------------------
sub _out {
    my($self, $value) = @_;
    if ($value =~ /^=/o) {
	return substr($value,1);
    } else {
	my $uri = $value;
	if ($self->{'dbqmap'}) {
#	    print STDERR "Before mapping: $uri\n";
	    $uri = $self->{'dbqmap'}->decode($uri);
#	    print STDERR "After mapping: $uri\n";
	}
	if ($self->{'extqmap'}) {
#	    print STDERR "Before mapping: $uri\n";
	    $uri = $self->{'extqmap'}->encode($uri);
#	    print STDERR "After mapping: $uri\n";
	}
	return $self->as_obj($uri);
    }
}

1;
